/*
 * spi.c
 *
 * SPI (M25P64) I/f
 *
 * Copyright (C) 2006-2008 Atmark Techno, Inc.	All Rights Reserved.
 *
 */

#include "memmap.h"

#ifdef SPI_REGISTER_BASEADDR

#include "medium.h"
#include "spi.h"

#define PRINT_CHECKSUM

/* SPI FIFO size */
#define USE_FIFO
#define FIFO_SIZE 16

/* Register */
#define SPI_INPUT_GLOBAL_ENABLE_REG	((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x1C)))
#define SPI_INTERRUPT_REG		((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x20)))
#define SPI_INTERRUPT_ENABLE_REG	((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x28)))
#if 0 /* unused */
#define SPI_RESET_MODULE		((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x40)))
#endif
#define SPI_CONTROL_REG			((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x060)))
#define SPI_STATUS_REG			((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x064)))
#define SPI_DATA_TRANSMIT_REG		((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x068)))
#define SPI_DATA_RECEIVE_REG		((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x06C)))
#define SPI_SLAVE_SELECT_REG		((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x070)))
#if 0 /* unused */
#define SPI_TRANSMIT_FIFO_OCCUPANCY	((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x074)))
#define SPI_RECEIVE_FIFO_OCCUPANCY	((*(volatile unsigned long *) (SPI_REGISTER_BASEADDR + 0x078)))
#endif

/* SPI Instruction set */
#define SPI_CODE_WREN		0x06 /* Write Enable */
#define SPI_CODE_WRDI		0x04 /* Write Disable */
#define SPI_CODE_RDID		0x9F /* Read Identification */
#define SPI_CODE_RDSR		0x05 /* Read Status Register */
#define SPI_CODE_WRSR		0x01 /* Write Status Register */
#define SPI_CODE_READ		0x03 /* Read Data Bytes */
#define SPI_CODE_FAST_READ	0x0B /* Read Data Bytes at Higher Speed */
#define SPI_CODE_PP		0x02 /* Page Program */
#define SPI_CODE_SE		0xD8 /* Sector Erase */
#define SPI_CODE_BE		0xC7 /* Bulk Erase */
#define SPI_CODE_RES		0xAB /* Read Electronic Signature */

/* SPI Control */
#define SPI_CONTROL_LOOP			0x001 /* Local loopback mode */
#define SPI_CONTROL_SPE				0x002 /* SPI System Enable select bit */
#define SPI_CONTROL_MSTR			0x004 /* MSTR */
#define SPI_CONTROL_CPOL			0x008 /* Clock Polarity select bit */
#define SPI_CONTROL_CPHA			0x010 /* Clocl Phase select bit */
#define SPI_CONTROL_TX_FIFO_RESET		0x020 /* Transmit FIFO Reset */
#define SPI_CONTROL_RX_FIFO_RESET		0x040 /* Receive FIFO Reset */
#define SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE	0x080 /* Manual SS Assertion Enable */
#define SPI_CONTROL_MASTER_TRANSATION_INHIBIT	0x100 /* Master Transaction Inhibit */

/* SPI Status */
#define SPI_STATUS_RX_EMPTY	0x01 /* Receive Empty */
#define SPI_STATUS_RX_FULL	0x02 /* Receive Full */
#define SPI_STATUS_TX_EMPTY	0x04 /* Transmit Empty */
#define SPI_STATUS_TX_FULL	0x08 /* Transmit Full */
#define SPI_STATUS_MODF		0x10 /* Mode-Fault Error Flag */

/* Reboot */
#define SYSTEM_GPIO_BASE	(SYSTEM_REGISTER_BASEADDR)
#define SYSTEM_GPIO(v)		(*(volatile unsigned long *)(SYSTEM_GPIO_BASE) = (v))
#define REBOOT			(1)

/* Functions */
static void SPI_Initialize(void);
static void SPI_WriteEnable(int cs);
static void SPI_WriteDisable(int cs);
static unsigned long SPI_ReadIdentification(int cs);
static unsigned char SPI_ReadStatusRegister(int cs);
#if 0 /* unused */
static void SPI_WriteStatusRegister(int cs);
#endif
static unsigned char SPI_ReadDataBytes(int cs, unsigned char addr);
static void SPI_FastReadDataBytes(int cs, unsigned long addr, void *buffer, int length);
static void SPI_PageProgram(int cs, unsigned long addr, void *data);
static void SPI_SectorErase(int cs,unsigned long addr);
#if 0 /* unused */
static void SPI_BulkErase(int cs, void);
static int SPI_ReadElectronicSignature(int cs, void);
#endif

static void busy_delay(int time)
{
	volatile int i; 
	for (i = 0; i < time; i++) {}
}

#ifdef PRINT_CHECKSUM
static void print_sum(unsigned long start, unsigned long size, char *lavel)
{
	volatile unsigned short *addr = (volatile unsigned short *)start;
	unsigned short temp;
	unsigned long sum = 0;

	for (; size >= 2; size -= 2) {
		temp = *addr++;
		sum += (temp >> 8) + (temp & 0xff);
	}
	if (size)
		sum += *(volatile unsigned char *)addr;

	myprint(lavel);
	myprint(" checksum: ");
	put_char_hex(sum);
}
#endif

/*
 *  SPI Copy from DRAM
 */
void spi_copy_from_dram(unsigned long start_addr, unsigned long last_addr)
{
	unsigned long offset_start, offset_last, offset;

	/* Initialize SPI */
	SPI_Initialize();

	/* Erase SPI */
	offset_start = ((start_addr) & 0x7F0000);
	offset_last = ((last_addr) & 0x7F0000);
	myprint("\r\nErasing SPI ...\r\n");

	/* Erase SPI */
	for (offset = offset_start; offset <= offset_last; offset += 0x10000)
		SPI_SectorErase(1, offset);

	/* Copy SDRAM to SPI */
	offset_start = ((start_addr) & 0x7FFF00);
	offset_last = ((last_addr) & 0x7FFF00);
	myprint("Programming SPI ...\r\n");

	/* Program SPI */
	for (offset = offset_start; offset <= offset_last; offset += 0x100)
		SPI_PageProgram(1, offset,
				(void *)(SDRAM_BASEADDR | (offset & 0x7FFFFF)));

	myprint("done.\r\n");

#ifdef PRINT_CHECKSUM
	{
		unsigned long start = SDRAM_BASEADDR + (start_addr & 0x7FFFFF);
		unsigned long size = (last_addr & 0x7FFFFF) -
			(start_addr & 0x7FFFFF) + 1;

		print_sum(start, size, "Original");

		spi_copy_to_dram(offset_start, (void *)start, size);
		print_sum(start, size, " Program");
	}
#endif

	myprint("Reboot.\r\n");
	busy_delay(1000000);
	SYSTEM_GPIO(REBOOT);
}

/*
 *  SPI Copy to DRAM
 */
void spi_copy_to_dram(unsigned long addr, void *buffer, int size)
{
	unsigned long offset;

	/* Initialize SPI */
	SPI_Initialize();

	/* Read SPI and Copy to DRAM */
	for (offset = 0x00000; offset < size; offset += 0x100)
		SPI_FastReadDataBytes(1, addr + offset,
				      (void *)(buffer + offset), 0x100);
}

unsigned long spi_read_identification(void)
{
	/* Initialize SPI */
	SPI_Initialize();

	return SPI_ReadIdentification(1);
}

/******************************************************************************
 *  SPI (M25P64) I/f
 */

/*
 *  Initialize
 */
void SPI_Initialize(void)
{
	/* Fifo reset Enable */
	SPI_CONTROL_REG =
		SPI_CONTROL_MASTER_TRANSATION_INHIBIT |
		SPI_CONTROL_RX_FIFO_RESET |
		SPI_CONTROL_TX_FIFO_RESET;

	/* Disable interrupt, and ack all to clear any bits */
	SPI_INPUT_GLOBAL_ENABLE_REG = 0x0;
	SPI_INTERRUPT_ENABLE_REG = 0x0;
	SPI_INTERRUPT_REG = SPI_INTERRUPT_REG;

	/* Deselect all slaves */
	SPI_SLAVE_SELECT_REG = 0xffffffff;

	/* Fifo reset Disable */
	SPI_CONTROL_REG =
		SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
		SPI_CONTROL_MSTR |
		SPI_CONTROL_SPE;
}

/*
 *  CS Enable
 */
static void SPI_CSEnable(int cs)
{
	SPI_SLAVE_SELECT_REG = ~cs;
	SPI_CONTROL_REG = SPI_CONTROL_MSTR | SPI_CONTROL_SPE;
	SPI_CONTROL_REG = SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
		SPI_CONTROL_MSTR | SPI_CONTROL_SPE;
}


/*
 *  CS Disable
 */
static void SPI_CSDisable(void)
{
	SPI_SLAVE_SELECT_REG = ~0;
	SPI_CONTROL_REG = SPI_CONTROL_MSTR | SPI_CONTROL_SPE;
	SPI_CONTROL_REG = SPI_CONTROL_MANUAL_SS_ASSERTION_ENABLE |
		SPI_CONTROL_MSTR | SPI_CONTROL_SPE;
}

/*
 *  OPB SPI Status TX FIFO empty check
 */
static inline int SPI_IsTxEmpty(void)
{
	return SPI_STATUS_REG & SPI_STATUS_TX_EMPTY;
}

/*
 *  OPB SPI Status TX FIFO full check
 */
static inline int SPI_IsTxFull(void)
{
	return SPI_STATUS_REG & SPI_STATUS_TX_FULL;
}

/*
 *  OPB SPI Status RX FIFO empty check
 */
static inline int SPI_IsRxEmpty(void)
{
	return SPI_STATUS_REG & SPI_STATUS_RX_EMPTY;
}

/*
 *  Write 8bit data
 */
static void SPI_Write8(const unsigned char v)
{
	int tmp;

	while (!SPI_IsTxEmpty());
	SPI_DATA_TRANSMIT_REG = v;
	while (SPI_IsRxEmpty());
	tmp = SPI_DATA_RECEIVE_REG;
}

/*
 *  Read 8bit data
 */
static unsigned char SPI_Read8(void)
{
	while (!SPI_IsTxEmpty());
	SPI_DATA_TRANSMIT_REG = 0x00;
	while (SPI_IsRxEmpty());
	return SPI_DATA_RECEIVE_REG;
}

#ifdef USE_FIFO

/*
 *  Write 24bit data
 */
static void SPI_Write24(const unsigned long data)
{
	int tmp;

	while (!SPI_IsTxEmpty());
	SPI_DATA_TRANSMIT_REG = (data >> 16) & 0xff;
	SPI_DATA_TRANSMIT_REG = (data >> 8) & 0xff;
	SPI_DATA_TRANSMIT_REG =	data & 0xff;
	while (SPI_IsRxEmpty());
	tmp = SPI_DATA_RECEIVE_REG;
	while (SPI_IsRxEmpty());
	tmp = SPI_DATA_RECEIVE_REG;
	while (SPI_IsRxEmpty());
	tmp = SPI_DATA_RECEIVE_REG;
}

/*
 *  Read 24bit data
 */
static unsigned long SPI_Read24(void)
{
	unsigned long tmp;

	while (!SPI_IsTxEmpty());
	SPI_DATA_TRANSMIT_REG = 0x00;
	SPI_DATA_TRANSMIT_REG = 0x00;
	SPI_DATA_TRANSMIT_REG = 0x00;
	while (SPI_IsRxEmpty());
	tmp = SPI_DATA_RECEIVE_REG << 16;
	while (SPI_IsRxEmpty());
	tmp |= SPI_DATA_RECEIVE_REG << 8;
	while (SPI_IsRxEmpty());
	return tmp | SPI_DATA_RECEIVE_REG	;
}

/*
 *  Write long data
 */
static void SPI_Write(const unsigned char *data, int length)
{
	int n, tmp;

	if (length >= FIFO_SIZE)
		n = FIFO_SIZE;
	else
		n = length;

	while (!SPI_IsTxEmpty());
	while (n--)
		SPI_DATA_TRANSMIT_REG = (unsigned long)*data++;

	for (; length > FIFO_SIZE; length--) {
		while (SPI_IsRxEmpty());
		tmp = SPI_DATA_RECEIVE_REG;
		while (SPI_IsTxFull());
		SPI_DATA_TRANSMIT_REG = (unsigned long)*data++;
	}

	while (length--) {
		while (SPI_IsRxEmpty());
		tmp = SPI_DATA_RECEIVE_REG;
	}
}

/*
 *  Read long data
 */
static void SPI_Read(unsigned char *data, int length)
{
	int n;

	if (length >= FIFO_SIZE)
		n = FIFO_SIZE;
	else
		n = length;

	while (!SPI_IsTxEmpty());
	while (n--)
		SPI_DATA_TRANSMIT_REG = 0x00;

	for (; length > FIFO_SIZE; length--) {
		while (SPI_IsRxEmpty());
		*data++ = (unsigned char)SPI_DATA_RECEIVE_REG;
		while (SPI_IsTxFull());
		SPI_DATA_TRANSMIT_REG = 0x00;
	}

	while (length--) {
		while (SPI_IsRxEmpty());
		*data++ = (unsigned char)SPI_DATA_RECEIVE_REG;
	}
}

#else

/*
 *  Write 24bit data
 */
static void SPI_Write24(const unsigned long data)
{
	SPI_Write8((data >> 16) & 0xff);
	SPI_Write8((data >> 8) & 0xff);
	SPI_Write8(data & 0xff);
}

/*
 *  Read 24bit data
 */
static unsigned long SPI_Read24(void)
{
	unsigned long tmp;

	tmp = SPI_Read8() << 16;
	tmp |= SPI_Read8() << 8;
	return tmp | SPI_Read8();
}

/*
 *  Write long data
 */
static void SPI_Write(const unsigned char *data, int length)
{
	while (length--)
		SPI_Write8(*data++);
}

/*
 *  Read long data
 */
static void SPI_Read(unsigned char *data, int length)
{
	while (length--)
		*data++ = SPI_Read8();
}

#endif

/*
 * Wait busy
 */
static inline void SPI_WaitBusy(int cs)
{
	while (SPI_ReadStatusRegister(cs) & 0x01);
}

/*
 *  Write Enable(SPI_CODE_WREN)
 */
void SPI_WriteEnable(int cs)
{
	SPI_CSEnable(cs);

	/* Write Enable Command */
	SPI_Write8(SPI_CODE_WREN);

	SPI_CSDisable();
}

/*
 *  Write Disable(SPI_CODE_WRDI)
 */
void SPI_WriteDisable(int cs)
{
	SPI_CSEnable(cs);

	/* Write Disable Command */
	SPI_Write8(SPI_CODE_WRDI);

	SPI_CSDisable();
}

/*
 *  Read Identification(SPI_CODE_RDID)
 */
unsigned long SPI_ReadIdentification(int cs)
{
	unsigned long idcode;

	SPI_CSEnable(cs);

	/* Read Identification Command */
	SPI_Write8(SPI_CODE_RDID);
	idcode = SPI_Read24();

	SPI_CSDisable();

	return idcode;
}

/*
 *  Read Status Register(SPI_CODE_RDSR)
 */
unsigned char SPI_ReadStatusRegister(int cs)
{
	unsigned char ret;

	SPI_CSEnable(cs);

	/* Read Status Register Command */
	SPI_Write8(SPI_CODE_RDSR);
	ret = SPI_Read8();

	SPI_CSDisable();

	return ret;
}

#if 0 /* unused */
/*
 *  Write Status Register(SPI_CODE_WRSR)
 */
void SPI_WriteStatusRegister(int cs)
{
}
#endif

/*
 *  Read Data Bytes(SPI_CODE_READ)
 */
unsigned char SPI_ReadDataBytes(int cs, unsigned char addr)
{
	unsigned char ret;

	SPI_CSEnable(cs);

	/* Read Status Register Command */
	SPI_Write8(SPI_CODE_READ);
	SPI_Write8(addr);
	ret = SPI_Read8();

	SPI_CSDisable();

	return ret;
}

/*
 *  Read Data Bytes at Higher Speed(SPI_CODE_FAST_READ)
 */
void SPI_FastReadDataBytes(int cs, unsigned long addr, void *buffer, int length)
{
	unsigned char *data = (unsigned char *)buffer ;

	SPI_CSEnable(cs);

	/* Read Command */
	SPI_Write8(SPI_CODE_FAST_READ);

	/* Addr */
	SPI_Write24(addr);

	/* Dummy write */
	SPI_Write8(0x00);

	/* Read data */
	SPI_Read(data, length);

	SPI_CSDisable();
}

/*
 *  Page Program(SPI_CODE_PP)
 */
void SPI_PageProgram(int cs, unsigned long addr, void *data)
{
	unsigned char *buffer = (unsigned char *)data;

	SPI_WriteEnable(cs);

	SPI_CSEnable(cs);

	/* Page Program Command */
	SPI_Write8(SPI_CODE_PP);

	/* Addr */
	SPI_Write24(addr);

	SPI_Write(buffer, 256);

	SPI_CSDisable();

	SPI_WaitBusy(cs);
}

/*
 *  Sector Erase(SPI_CODE_SE)
 */
void SPI_SectorErase(int cs, unsigned long addr)
{
	SPI_WriteEnable(cs);

	SPI_CSEnable(cs);

	/* Sector Erase Comand */
	SPI_Write8(SPI_CODE_SE);

	/* Addr */
	SPI_Write24(addr);

	SPI_CSDisable();

	SPI_WaitBusy(cs);
}

#if 0 /* unused */
/*
 *  Bulk Erase(SPI_CODE_BE)
 */
void SPI_BulkErase(int cs)
{
	SPI_WriteEnable(cs);

	SPI_CSEnable(cs);

	/* Write Enable Command */
	SPI_Write8(SPI_CODE_BE);

	SPI_CSDisable();

	SPI_WaitBusy(cs);
}

/*
 *  Read Electronic Signature(SPI_CODE_RES)
 */
int SPI_ReadElectronicSignature(int cs)
{
	return -1;
}
#endif

#endif /* SPI_REGISTER_BASEADDR */
